@@ -1,7 +1,7 @@
 package Cache::Memcached::AnyEvent::Protocol::Binary;
 use strict;
 use base 'Cache::Memcached::AnyEvent::Protocol';
-use bytes;
+use Carp qw(confess);
 use constant HEADER_SIZE => 24;
 use constant HAS_64BIT => do {
     no strict;
@@ -148,42 +148,32 @@ AnyEvent::Handle::register_read_type memcached_bin => sub {
 
     my %state = ( waiting_header => 1 );
     sub {
-        return unless $_[0]{rbuf};
-
-        my $rbuf_ref = \$_[0]{rbuf};
         if ($state{waiting_header}) {
-            return if length $$rbuf_ref < HEADER_SIZE;
-
-            my $header = substr $$rbuf_ref, 0, HEADER_SIZE, '';
-            my ($i1, $i2, $i3, $i4, $i5, $i6) = unpack('N6', $header);
-            $state{magic}             = $i1 >> 24;
-            $state{opcode}            = ($i1 & 0x00ff0000) >> 16;
-            $state{key_length}        = ($i1 & 0x0000ffff);
-            $state{extra_length}      = ($i2 & 0xff000000) >> 24;
-            $state{data_type}         = ($i2 & 0x00ff0000) >> 8;
-            $state{status}            = ($i2 & 0x0000ffff);
-            $state{total_body_length} = $i3;
-            $state{opaque}            = $i4;
-
-            if (HAS_64BIT) {
-                $state{cas} = $i5 << 32 + $i6;
-            } else {
-                warn "overflow on CAS" if ($i5 || 0) != 0;
-                $state{cas} = $i6;
-            }
+            bytes::length ( $_[0]{rbuf} || '') >= HEADER_SIZE or return;
+            my $header = bytes::substr $_[0]{rbuf}, 0, HEADER_SIZE, "";
+            my ($magic, $opcode, $key_length, $extra_length, $status, $data_type, $total_body_length, $opaque, $cas) = _decode_header($header);
+            $state{magic} = $magic;
+            $state{opcode} = $opcode;
+            $state{key_length} = $key_length;
+            $state{extra_length} = $extra_length;
+            $state{status} = $status;
+            $state{data_type} = $data_type;
+            $state{total_body_length} = $total_body_length;
+            $state{opaque} = $opaque;
+            $state{cas} = $cas;
 
             delete $state{waiting_header};
         }
 
         if ($state{total_body_length}) {
-            return if length $$rbuf_ref < $state{total_body_length};
+            bytes::length $_[0]{rbuf} >= $state{total_body_length} or return;
 
-            $state{extra} = substr $$rbuf_ref, 0, $state{extra_length}, '';
-            $state{key} = substr $$rbuf_ref, 0, $state{key_length}, '';
+            $state{extra} = bytes::substr $_[0]{rbuf}, 0, $state{extra_length}, '';
+            $state{key} = bytes::substr $_[0]{rbuf}, 0, $state{key_length}, '';
 
 
             my $value_len = $state{total_body_length} - ($state{key_length} + $state{extra_length});
-            $state{value} = substr $$rbuf_ref, 0, $value_len, '';
+            $state{value} = bytes::substr $_[0]{rbuf}, 0, $value_len, '';
         }
 
         $cb->( \%state );
@@ -198,8 +188,16 @@ sub prepare_handle {
 }
 
 AnyEvent::Handle::register_write_type memcached_bin => sub {
-    my ($self, $opcode, $key, $extras, $body, $cas, $data_type, $reserved ) = @_;
-    my $key_length = defined $key ? length($key) : 0;
+    my ($self, @args) = @_;
+    return _encode_request(@args);
+};
+
+sub _encode_request {
+    my ( $opcode, $key, $extras, $body, $cas, $data_type, $reserved ) = @_;
+
+    use bytes;
+
+    my $key_length = defined $key ? bytes::length($key) : 0;
     # first 4 bytes (long)
     my $i1 = 0;
     $i1 ^= REQ_MAGIC << 24;
@@ -207,19 +205,13 @@ AnyEvent::Handle::register_write_type memcached_bin => sub {
     $i1 ^= $key_length;
 
     # second 4 bytes
+    my $extra_length = defined $extras ? bytes::length($extras) : 0;
     my $i2 = 0;
-    my $extra_length = 
-        ($opcode != MEMD_PREPEND && $opcode != MEMD_APPEND && defined $extras) ?
-        length($extras) :
-        0
-    ;
-    if ($extra_length) {
-        $i2 ^= $extra_length << 24;
-    }
+    $i2 ^= $extra_length << 24;
     # $data_type and $reserved are not used currently
 
     # third 4 bytes
-    my $body_length  = defined $body ? length($body) : 0;
+    my $body_length  = defined $body ? bytes::length($body) : 0;
     my $i3 = $body_length + $key_length + $extra_length;
 
     # this is the opaque value, which will be returned with the response
@@ -245,8 +237,8 @@ AnyEvent::Handle::register_write_type memcached_bin => sub {
     }
 
     my $message = pack( 'N6', $i1, $i2, $i3, $i4, $i5, $i6 );
-    if (length($message) > HEADER_SIZE) {
-        Carp::confess "header size assertion failed";
+    if (bytes::length($message) > HEADER_SIZE) {
+        confess "header size assertion failed";
     }
 
     if ($extra_length) {
@@ -262,6 +254,33 @@ AnyEvent::Handle::register_write_type memcached_bin => sub {
     return $message;
 };
 
+use constant _noop => _encode_request(MEMD_NOOP, undef, undef, undef, undef, undef, undef);
+
+sub _decode_header {
+    my $header = shift;
+
+    my ($i1, $i2, $i3, $i4, $i5, $i6) = unpack('N6', $header);
+    my $magic = $i1 >> 24;
+    my $opcode = ($i1 & 0x00ff0000) >> 16;
+    my $key_length = $i1 & 0x0000ffff;
+    my $extra_length = ($i2 & 0xff000000) >> 24;
+    my $data_type = ($i2 & 0x00ff0000) >> 8;
+    my $status = $i2 & 0x0000ffff;
+    my $total_body_length = $i3;
+    my $opaque = $i4;
+
+    my $cas;
+    if (HAS_64BIT) {
+        $cas = $i5 << 32;
+        $cas += $i6;
+    } else {
+        warn "overflow on CAS" if ($i5 || 0) != 0;
+        $cas = $i6;
+    }
+
+    return ($magic, $opcode, $key_length, $extra_length, $status, $data_type, $total_body_length, $opaque, $cas);
+}
+
 sub _status_str {
     my $status = shift;
     my %strings = (
@@ -278,20 +297,25 @@ sub _status_str {
 
 # Generate setters
 {
+    my %bincmd = (
+        add => MEMD_ADD(),
+        replace => MEMD_REPLACE(),
+        set => MEMD_SET(),
+    );
     my $generator = sub {
-        my ($cmd, $opcode) = @_;
+        my $cmd = shift;
 
+        my $bincmd = $bincmd{$cmd};
         sub {
-            my ($self, $guard, $memcached, $key, $value, $expires, $noreply, $cb) = @_;
-            my $fq_key = $memcached->_prepare_key( $key );
-            my $handle = $memcached->_get_handle_for( $fq_key );
-            my ($len, $flags);
+            my ($guard, $self, $memcached, $key, $value, $exptime, $noreply, $cb) = @_;
+            my $handle = $self->get_handle_for( $key );
 
-            $memcached->_prepare_value( $cmd, \$value, \$len, \$expires, \$flags);
+            my ($write_data, $write_len, $flags, $expires) =
+                $self->prepare_value( $cmd, $value, $exptime || 0);
 
             my $extras = pack('N2', $flags, $expires);
 
-            $handle->push_write( memcached_bin => $opcode, $fq_key, $extras, $value );
+            $handle->push_write( memcached_bin => $bincmd, $key, $extras, $write_data );
             $handle->push_read( memcached_bin => sub {
                 undef $guard;
                 $cb->($_[0]->{status} == 0, $_[0]->{value}, $_[0]);
@@ -299,100 +323,103 @@ sub _status_str {
         }
     };
 
-    *add     = $generator->("add", MEMD_ADD);
-    *replace = $generator->("replace", MEMD_REPLACE);
-    *set     = $generator->("set", MEMD_SET);
-    *append  = $generator->("append", MEMD_APPEND);
-    *prepend = $generator->("prepend", MEMD_PREPEND);
+    sub _build_add_cb {
+        my $self = shift;
+        return $generator->("add");
+    }
+
+    sub _build_replace_cb {
+        my $self = shift;
+        return $generator->("replace");
+    }
+
+    sub _build_set_cb {
+        my $self = shift;
+        return $generator->("set");
+    }
 }
 
-sub delete {
-    my ($self, $guard, $memcached, $key, $noreply, $cb) = @_;
+sub _build_delete_cb {
+    return sub {
+        my ($guard, $self, $memcached, $key, $noreply, $cb) = @_;
 
-    my $fq_key = $memcached->_prepare_key($key);
-    my $handle = $memcached->_get_handle_for($fq_key);
+        my $handle = $self->get_handle_for($key);
 
-    $handle->push_write( memcached_bin => MEMD_DELETE, $fq_key );
-    $handle->push_read( memcached_bin => sub {
-        undef $guard;
-        $cb->(@_);
-    } );
-}
-
-sub get {
-    my ($self, $guard, $memcached, $key, $cb) = @_;
-
-    my $fq_key = $memcached->_prepare_key( $key );
-    my $handle = $memcached->_get_handle_for( $fq_key );
-    $handle->push_write(memcached_bin => MEMD_GETK, $fq_key);
-    $handle->push_read(memcached_bin => sub {
-        my $msg = shift;
-        my ($flags, $exptime) = unpack('N2', $msg->{extra});
-        if (exists $msg->{key} && exists $msg->{value}) {
-            my $value = $msg->{value};
-            $memcached->_decode_key_value(\$key, \$flags, \$value );
+        $handle->push_write( memcached_bin => MEMD_DELETE, $key );
+        $handle->push_read( memcached_bin => sub {
             undef $guard;
-            $cb->($value);
-        }
-    });
+            $cb->(@_);
+        } );
+    }
 }
 
-sub get_multi {
-    my ($self, $guard, $memcached, $keys, $cb, $cb_caller) = @_;
+sub _build_get_multi_cb {
+    return sub {
+        my ($guard, $self, $memcached, $type, $keys, $cb, $cb_caller) = @_;
 
-    # organize the keys by handle
-    my %handle2keys;
-
-    foreach my $key (@$keys) {
-        my $fq_key = $memcached->_prepare_key( $key );
-        my $handle = $memcached->_get_handle_for( $fq_key );
-        my $list = $handle2keys{ $handle };
-        if (! $list) {
-            $handle2keys{$handle} = [ $handle, $fq_key ];
-        } else {
-            push @$list, $fq_key;
+        # organize the keys by handle
+        my %handle2keys;
+            
+        foreach my $key (@$keys) {
+            my $handle = $self->get_handle_for( $key );
+            my $list = $handle2keys{ $handle };
+            if (! $list) {
+                $handle2keys{$handle} = [ $handle, $key ];
+            } else {
+                push @$list, $key;
+            }
         }
-    }
 
-    my $cv = AE::cv {
-        undef $guard;
-        $cb->( $_[0]->recv );
-    };
-
-    my %result;
-    $cv->begin( sub { $_[0]->send(\%result) } );
-    foreach my $list (values %handle2keys) {
-        my ($handle, @keys) = @$list;
-        foreach my $key ( @keys ) {
-            $handle->push_write(memcached_bin => MEMD_GETK, $key);
-            $cv->begin;
-            $handle->push_read(memcached_bin => sub {
-                my $msg = shift;
-
-                my ($flags, $exptime) = unpack('N2', $msg->{extra});
-                if (exists $msg->{key} && exists $msg->{value}) {
-                    my $value = $msg->{value};
-                    $memcached->_decode_key_value(\$key, \$flags, \$value );
-                    $result{ $key } = $value;
+        my $cv = $type eq 'single' ?
+            AE::cv {
+                undef $guard;
+                my ($rv, $msg) = $_[0]->recv;
+                if ($rv) {
+                    ($rv) = values %$rv;
                 }
-                $cv->end;
-            });
+                $cb->($rv, $msg);
+            } :
+            AE::cv {
+                undef $guard;
+                $cb->( $_[0]->recv );
+            }
+        ;
+
+        my %result;
+        $cv->begin( sub { $_[0]->send(\%result) } );
+        foreach my $list (values %handle2keys) {
+            my ($handle, @keys) = @$list;
+            foreach my $key ( @keys ) {
+                $handle->push_write(memcached_bin => MEMD_GETK, $key);
+                $cv->begin;
+                $handle->push_read(memcached_bin => sub {
+                    my $msg = shift;
+
+                    my ($flags, $exptime) = unpack('N2', $msg->{extra});
+                    if (exists $msg->{key} && exists $msg->{value}) {
+                        $result{ $msg->{key} } = exists $msg->{value} ? $msg->{value} : undef;
+                    }
+                    $cv->end;
+                });
+
+                $handle->push_write( _noop() );
+                $handle->push_read( memcached_bin => sub {} );
+            }
         }
-    }
-    $cv->end;
+        $cv->end;
+    };
 }
 
 {
     my $generator = sub {
         my ($opcode) = @_;
         return sub {
-            my ($self, $guard, $memcached, $key, $value, $initial, $cb) = @_;
+            my ($guard, $self, $memcached, $key, $value, $initial, $cb) = @_;
     
             $value ||= 1;
             my $expires = defined $initial ? 0 : 0xffffffff;
             $initial ||= 0;
-            my $fq_key = $memcached->_prepare_key( $key );
-            my $handle = $memcached->_get_handle_for($fq_key);
+            my $handle = $self->get_handle_for($key);
             my $extras;
             if (HAS_64BIT) {
                 $extras = pack('Q2L', $value, $initial, $expires );
@@ -401,7 +428,7 @@ sub get_multi {
             }
     
             $handle->push_write(memcached_bin => 
-                $opcode, $fq_key, $extras, undef, undef, undef, undef);
+                $opcode, $key, $extras, undef, undef, undef, undef);
             $handle->push_read(memcached_bin => sub {
                 undef $guard;
                 my $value;
@@ -416,91 +443,35 @@ sub get_multi {
         }
     };
 
-    *incr = $generator->(MEMD_INCREMENT);
-    *decr = $generator->(MEMD_DECREMENT);
+    sub _build_incr_cb {
+        return $generator->(MEMD_INCREMENT);
+    }
+
+    sub _build_decr_cb {
+        return $generator->(MEMD_DECREMENT);
+    }
 };
 
-sub version {
-    my ($self, $guard,$memcached, $cb) = @_;
+sub _build_version_cb {
+    return sub {
+        my ($guard, $self, $memcached, $cb) = @_;
 
-    my %ret;
-    my $cv = AE::cv { $cb->( \%ret ); undef %ret };
-    while (my ($host_port, $handle) = each %{ $memcached->{_server_handles} }) {
-        $handle->push_write(memcached_bin => MEMD_VERSION);
-        $cv->begin;
-        $handle->push_read(memcached_bin => sub {
-            my $msg = shift;
-            undef $guard;
-            my $value = unpack('a*', $msg->{value});
+        my %ret;
+        my $cv = AE::cv { $cb->( \%ret ); undef %ret };
+        while (my ($host_port, $handle) = each %{ $memcached->{_server_handles} }) {
+            $handle->push_write(memcached_bin => MEMD_VERSION);
+            $cv->begin;
+            $handle->push_read(memcached_bin => sub {
+                my $msg = shift;
+                undef $guard;
+                my $value = unpack('a', $msg->{value});
 
-            $ret{ $host_port } = $value;
-            $cv->end;
-        });
+                $ret{ $host_port } = $value;
+                $cv->end;
+            });
+        }
     }
 }
         
-sub flush_all {
-    my ($self, $guard, $memcached, $delay, $noreply, $cb) = @_;
-
-    my $cv = AE::cv {
-        undef $guard;
-        $cb->(1);
-    };
-
-    while (my ($host_port, $handle) = each %{ $memcached->{_server_handles} }) {
-        $handle->push_write(memcached_bin => MEMD_FLUSH);
-        $cv->begin;
-        $handle->push_read(memcached_bin => sub {
-            $cv->end;
-        });
-    }
-}
-
-1;
-
-__END__
-
-=head1 NAME
-
-Cache::Memcached::AnyEvent::Protocol::Binary - Implements Memcached Binary Protocol
-
-=head1 SYNOPSIS
-
-    use Cache::Memcached::AnyEvent;
-    my $memd = Cache::Memcached::AnyEvent->new({
-        ...
-        protocol_class => 'Binary'
-    });
-
-=head1 METHODS
-
-=head2 add
-
-=head2 append
-
-=head2 decr
-
-=head2 delete
-
-=head2 flush_all
-
-=head2 get
-
-=head2 get_multi
-
-=head2 incr
-
-=head2 prepare_handle
-
-=head2 prepend
-
-=head2 replace
-
-=head2 set
-
-=head2 stats
-
-=head2 version
-
-=cut
 
+1;
\ No newline at end of file
